/*
* ==============================================================================
*  Name        : ShoutcastStream.cpp
*  Part of     : Shoutcast Engine
*  Interface   :
*  Description : Implementation of shoutcast stream class
*  Version     : 2
*
*  Copyright (c) 2006, Nokia Corporation All rights reserved. Redistribution
*  and use in source and binary forms, with or without modification, are
*  permitted provided that the following conditions are met: Redistributions
*  of source code must retain the above copyright notice, this list of
*  conditions and the following disclaimer. Redistributions in binary form
*  must reproduce the above copyright notice, this list of conditions and the
*  following disclaimer in the documentation and/or other materials provided
*  with the distribution. Neither the name of the Nokia Corporation nor the
*  names of its contributors may be used to endorse or promote products
*  derived from this software without specific prior written permission. THIS
*  SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
*  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
* ==============================================================================
*/

// INCLUDE FILES
#include <e32base.h>
#include <f32file.h>

#include <libc/stdlib.h>
#include <libc/string.h>
#include <libc/sys/unistd.h> //sleep
#include <errno.h>
#include <assert.h>

#include "ShoutcastDefs.h"
#include "ShoutcastErrors.h"
#include "ShoutcastUIDs.hrh"
#include "ShoutcastStream.h"
#include "ShoutcastEventDispatcher.h"

// CONSTANTS


// Formating strings
_LIT(KSCBufFormat, "Buf:%3d");
_LIT(KSCTechDataFormat, "%dkb/s %dkHz ");
_LIT(KSCDataRcvdFormat, "%3.1fMB");
_LIT(KSCMetaPriceFormat, "price: %3.1f");
_LIT(KSCMetaBitrateFormat, "Link:%d/%d");

// Audio formats
_LIT(KSCMono, "Mono");
_LIT(KSCStereo, "Stereo");
_LIT(KSCMp3, " MP3");
_LIT(KSCAacp, " AAC+");

// HTTP Request
_LIT8(KSCGet, "GET ");
_LIT8(KSCHttp10, " HTTP/1.0\r\n");
_LIT8(KSCHost, "Host: ");
_LIT8(KSCUserAgent, "User-Agent:S60InternetRadio/2.0\r\n");
_LIT8(KSCAccept, "Accept:*/*\r\n");
_LIT8(KSCIcyMetadata, "icy-metadata:1\r\n");
_LIT8(KSCConnectionClose, "Connection:close\r\n\r\n");

// HTTP Response
_LIT8(KSCOKHeader, "200 OK");
_LIT8(KSCContentTypeTag, "content-type");
_LIT(KSCMimeTypeAudioMpeg, "audio/mpeg");
_LIT(KSCMimeTypeAudioAacp, "audio/aacp");
_LIT(KSCMimeTypeAudioAac, "audio/aac");
_LIT8(KSCIcyMetaintTag, "icy-metaint");
_LIT8(KSCIcyNameTag, "icy-name");
_LIT8(KSCIcyGenreTag, "icy-genre");
_LIT8(KSCIcyBrTag, "icy-br");

// Internal metadata names
_LIT(KSCMetaServer, "Server");
_LIT(KSCMetaTechDetails, "TechDetails");
_LIT(KSCMetaArtistTitle, "ArtistTitle");
_LIT(KSCMetaGenre, "Genre");
_LIT(KSCMetaBytes, "Bytes");
_LIT(KSCMetaPrice, "Price");
_LIT(KSCMetaBuffer, "Buffer");
_LIT(KSCMetaBitrate, "Bitrate");

_LIT8(KSCStreamTitle, "StreamTitle='");

const TInt K100kBFactor = 104857;
const TReal K1MBFactor = 1048576.0;

// Bit rates in bits/sec supported by MPEG2, MPEG1 and MPEG2.5 respectively
const TInt16 KMp3BitRateTable[3][16] =
	{
		{-1,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},
		{-1,32,40,48,56,64,80,96,112,128,160,192,224,256,320,0},
		{-1,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},
	};

// Sampling frequencies supported by MPEG2, MPEG1 and MPEG2.5 respectively
const TUint16 KMp3SamplingFrequencyTable[3][4] =
	{
		{22050,24000,16000,0},
		{44100,48000,32000,0},
		{11025,12000,8000,0},
	};

const TUint KAacSamplingFrequencyTable[16] =
	{
	96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,0,0,0,0
	};

// Maximum MP3 frame size
const TInt KMp3MaxFrameSize = 1440; // 320kbit/s @ 32kHz

// The size of MP3 header, header must include bits for determining frame length
const TInt KMp3FrameHeaderSize = 5;
const TInt KAacFrameHeaderSize = 7;

// ============================ MEMBER FUNCTIONS ===============================

// -----------------------------------------------------------------------------
// CShoutcastStream::CShoutcastStream
// C++ default constructor can NOT contain any code, that might leave.
// -----------------------------------------------------------------------------
//
CShoutcastStream::CShoutcastStream()
	: CActive(EPriorityStandard),
	  iState(ENotConnected),
	  iDispatcher(NULL),
	  iStreamOutput(NULL),
	  iPtrBuffer2Decode(NULL,0,0),
	  iPtrBuffer2Play(NULL,0,0),
	  iMaxFrameSize(0),
	  iVolume(0),
	  iMetadata(NR_METADATA_FIELDS)
    {
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::NewL
// Two-phased constructor.
// -----------------------------------------------------------------------------
//
CShoutcastStream* CShoutcastStream::NewL(
	const TDesC8& aUrl,
	MShoutcastStreamObserver& aObserver )
    {
    CShoutcastStream* self = new (ELeave) CShoutcastStream();
    CleanupStack::PushL(self);
    self->ConstructL(aUrl, aObserver);
	CleanupStack::Pop(self);
    return self;
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::ConstructL
// Symbian 2nd phase constructor can leave.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ConstructL(
	const TDesC8& aUrl,
	MShoutcastStreamObserver& aObserver )
	{
	LOG_START;

	LOG("CShoutcastStream::ConstructL");

	iURL.Copy(aUrl);
	iTempMetadata.Copy(iURL);
	LOG1("CShoutcastStream::OpenL: %S",&iTempMetadata);

	//here gotUrl should be true and iURL should be filled!
	GetIPFromURL();

	iDispatcher = CEventDispatcher::NewL(aObserver);
	// Defaults to MP3 encoded stream
	iDataType.Set(KMMFFourCCCodeMP3);

	CActiveScheduler::Add(this);
	LOG("CShoutcastStream::ConstructL OK");

	}


// Destructor
CShoutcastStream::~CShoutcastStream()
    {
	LOG("CShoutcastStream::~CShoutcastStream");
	Cancel();
	CloseAndClean();

	delete iAddr;iAddr=NULL;

	if ( iStreamOutput )
		{
		delete iStreamOutput;
		iStreamOutput = NULL;
		}

	delete iDispatcher;

	LOG("CShoutcastStream Destructor OK");
	LOG_END;

	}


// -----------------------------------------------------------------------------
// CShoutcastStream::CloseAndClean
// Close connections and clean up resources
// -----------------------------------------------------------------------------
//

void CShoutcastStream::CloseAndClean()
	{

	if ( iState == ENotConnected )
		return;

	iState = EDisconnecting;

	//buffering stuff
	iPtrBuffer2Decode.Set(NULL,0,0);
	iPtrBuffer2Play.Set(NULL,0,0);
	delete iBuffer2Decode;
	iBuffer2Decode = NULL;
	iLenBuffer2Decode = iPosBuffer2Decode=0;
	iLenBuffer2Play = 0;
	delete iBuffer2Play;
	iBuffer2Play = NULL;

	//close connection
	iSock.Close();
	iSocksvr.Close();

	//metadata
	iMetadata.ResetAndDestroy();
	delete iMetadataBuffer;
	iMetadataBuffer = NULL;

	iState = ENotConnected;
	}




// -----------------------------------------------------------------------------
// CShoutcastStream::MaoscBufferCopied
// The descriptor buffer has been copied, fill it again with encoded data.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::MaoscBufferCopied(
	TInt aError,
	const TDesC8& /*aBuffer*/)
	{
	LOG1("MaoscBufferCopied: iState = %d ",iState);
	ASSERT(iState == EData);
	iWritingToStream = EFalse;
	// if no error, write more data to the output stream
	if ( !aError )
		{

		TRAPD(err, FillBufferL());
		if ( err )
			{
			iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),err);
			}
		}
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::FillBuffer
// Fill the play buffer with encoded data and send it to the stream output.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::FillBufferL()
	{
	LOG1("FillBufferL: iState = %d", iState);
	ASSERT(iState == EData);

	if ( iWritingToStream )
		{
		// Still busy writing to stream. We have not received MaoscBufferCopied.
		// Bailing out...
		return;
		}
	else
		{
		iWritingToStream = ETrue;
		}

	TInt ldTotal;

	iLenBuffer2Play=iPtrBuffer2Play.MaxLength();
	if ( iLenBuffer2Decode+iReadData < iLenBuffer2Play )
		{
		LOG("FillBufferL: Not enough data to play. Rebuffering");
		//not enough data to play
		iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),KShoutcastEvent_BufferEmpty);
		iLenBuffer2Play=0;
		iPausedForBuffering = ETrue;
		iWritingToStream = EFalse;
		}
	else
		{
		//there is enough data to be played!
		if ( iLenBuffer2Decode >= iLenBuffer2Play )
			{
			//copy all the data into the buffer
			iPtrBuffer2Play.Copy(iBuffer2Decode+iPosBuffer2Decode,iLenBuffer2Play);
			iPosBuffer2Decode+=iLenBuffer2Play;
			iLenBuffer2Decode-=iLenBuffer2Play;
			}
		else
			{
			//we have to copy the entire end of the buffer, then also the beginning of the buffer
			//copy the end of the buffer
			iPtrBuffer2Play.Copy(iBuffer2Decode+iPosBuffer2Decode, iLenBuffer2Decode);
			TInt dataLeft = iLenBuffer2Play-iLenBuffer2Decode;
			//we need to go from case 2 to case 1 (when there is no buffer to copy)
			ASSERT(iReadData > 0);
			iPosBuffer2Decode=0;
			iLenBuffer2Decode=iReadData;
			iReadData=-1;//going from case 2 to case 1
			//append the remaining data
			ASSERT(dataLeft <= iLenBuffer2Decode);//due to the first if from the HwDevice else
			iPtrBuffer2Play.Append(iBuffer2Decode+iPosBuffer2Decode,dataLeft);
			iPosBuffer2Decode+=dataLeft;
			iLenBuffer2Decode-=dataLeft;
			}
		}
	ASSERT(iLenBuffer2Play >= 0);
	ldTotal = iLenBuffer2Play;

	if ( iLenBuffer2Play > 0 )
		{
		LOG1("FillBufferL: Calling WriteL for %d bytes",iLenBuffer2Play);
		iStreamOutput->WriteL(iPtrBuffer2Play);
		}


	//buffering stuff
	if(iPosBuffer2Decode>BUFFER2DECODE_SIZE-iMaxFrameSize-READ_EPSILON && iLenBuffer2Decode<iMaxFrameSize)
	{
		//we passed the upper safety limit!
		LOG2("FillBufferL: Upper limit passed (pos=%d, len=%d). ",iPosBuffer2Decode,iLenBuffer2Decode);
		ASSERT(iLenBuffer2Decode < iMaxFrameSize);
		ASSERT(iReadData >= 0);

		//copy the upper part somewhere down
		memmove(iBuffer2Decode+iMaxFrameSize-iLenBuffer2Decode,iBuffer2Decode+iPosBuffer2Decode,iLenBuffer2Decode);
		iPosBuffer2Decode=iMaxFrameSize-iLenBuffer2Decode;
		iLenBuffer2Decode+=iReadData;
		iReadData=-1;//going from case 2 to case 1
	};

	//buffering fill
	TInt bufferFill;
	if(iReadData==-1)
		{
		bufferFill=(TInt)(100.0*iLenBuffer2Decode/(BUFFER2DECODE_SIZE-iMaxFrameSize));
		}
	else
		{
		bufferFill=(TInt)(100.0*(iLenBuffer2Decode+iReadData)/(BUFFER2DECODE_SIZE-iMaxFrameSize));
		}

	//fill buffering metadata
	iTempMetadata.Format(KSCBufFormat, bufferFill);
	iMetadata[6]->SetValueL(iTempMetadata);
	//update metadata on the client/controller
	if ( !iPausedForBuffering )
		{
		iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),0x0140);
		}

	//restarting buffering, if needed!
	if(!iReadingActive && ldTotal>0)
	{
		//reading is stopped and we need to restart it!
		ReadRequest();
	};

	LOG2("FillBufferL OK (pos=%d, len=%d)",iPosBuffer2Decode, iLenBuffer2Decode);

	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ResetBufferVars
// Reset buffer variables
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ResetBufferVars()
	{
    iPosBuffer2Decode=0;
	iLenBuffer2Decode=0;
	iReadData=-1;
	iReadingActive=EFalse;
	iPtrBuffer2Decode.Set(iBuffer2Decode,0,BUFFER2DECODE_SIZE);
	iPtrBuffer2Play.Set(iBuffer2Play,0,BUFFER2PLAY_SIZE);
	iBitsRead.Reset();
	iTimeIntervals.Reset();
	iLastReading=0;
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::InitStreamOutputL
// Create and initialize the output stream  This must be done before playing can
// began.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::InitStreamOutputL()
	{
	LOG("CShoutcastStream::InitStreamOutputL");
	ASSERT(iState == EInitStreamOutput);
	iStreamOutput = CMdaAudioOutputStream::NewL(*this);
//		iStreamOutput->SetPriority(EMdaPriorityMin , EMdaPriorityPreferenceTimeAndQuality);
	iStreamOutput->SetPriority(80 , EMdaPriorityPreferenceTimeAndQuality);
	iStreamOutput->Open(&iSettings);
	LOG("CShoutcastStream::InitStreamOutputL OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::MaoscOpenComplete
// Once the stream is opened, we can set audio properties, volume, and priority.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::MaoscOpenComplete(
	TInt aError )
	{

	LOG("MaoscOpenComplete");

	TInt err;
	if ( aError == KErrNone )
		{

		// If this leaves with KErrNotSupported, no decoder was found that support the
		// specified encoded data stream.
		TRAP(err, iStreamOutput->SetDataTypeL(iDataType.FourCC()));
		if ( err )
			{
			if ( err == KErrNotSupported )
				{
				iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), KErrShoutcast_FormatNotSupported);
				}
			else
				{
				iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), err);
				}
			LOG1("SetDataType: Leaving with %d err",err);
			return;
			}

        // [Petteri] This needs to be set after the data type, at least on
        // 9500 and 6630, otherwise the output will be 8kHz mono or something
        // similar
		TRAP(err, iStreamOutput->SetAudioPropertiesL(iAudioSettings.iSampleRate, iAudioSettings.iChannels));
		if ( err )
			{
			iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), err);
			LOG1("SetAudioPropertiesL: Leaving with %d err", err);
			return;
			}

		LOG1("SetAudioPropertiesL: Max Volume = %d", iStreamOutput->MaxVolume());
		iStreamOutput->SetVolume(iVolume);

		//announce the client that we are connected
		iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),KShoutcastEvent_Connected);
		iState = EData;//this should be the only place where iState gets assigned the EData value!
		}
	else
		{
		iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), aError);
		}

	LOG("MaoscOpenComplete OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::MaoscPlayComplete
// Called when the end of the sound data has been reached or the
// playing has stopped for some other reason.
// KErrUnderflow = "end of the sound data has been reached"
// -----------------------------------------------------------------------------
//
void CShoutcastStream::MaoscPlayComplete(
	TInt aError )
	{
	LOG1("MaoscPlayComplete: err=%d",aError);

	if ( aError == KErrUnderflow )
		{
		// This case where we run out of data to play and needs to rebuffer.
		iPausedForBuffering = ETrue;
		}
	if ( aError == KErrDied || aError == KErrAccessDenied )
		{
		Stop();
		iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), aError);
		}

	}


// -----------------------------------------------------------------------------
// CShoutcastStream::ConnectToServerL
// Attempt to open and establish connection with the server.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ConnectToServerL(
	TBool aReconnecting )
	{
	LOG("CShoutcastStream::ConnectToServerL");
	TInt err;

	//initialize data related to networking
	ASSERT(iState == ENotConnected);

	if ( !aReconnecting )
		{
		//retrieving the address and connecting to Socket Server can be done only once!
		err = iSocksvr.Connect();
		if ( err != KErrNone )
			{
			User::Leave(KErrShoutcast_ConnectToSockServer+err);
			}
		LOG("ConnectToServerL: Connected to ESOCK server");

		//opening the socket can also be done only once
		err = iSock.Open(iSocksvr, KAfInet, KSockStream, KProtocolInetTcp);
		LOG1("ConnectToServerL: Client sock opened, err=%d",err);
		if( err != KErrNone )
			{
			User::Leave(KErrShoutcast_OpenSocket+err);
			}
		}

	//announce to the client that there will be a lengthly operation
	iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),
			aReconnecting?KShoutcastEvent_Reconnecting:KShoutcastEvent_Connecting);

	//connecting
	iState = EConnecting;
	iSock.Connect(*iAddr, iStatus);
	SetActive();
	LOG("CShoutcastStream::ConnectToServerL OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::SendRequestToServerL
// Sends HTTP requests to the server
// -----------------------------------------------------------------------------
//
void CShoutcastStream::SendRequestToServerL()
	{
	LOG1("SendRequestToServerL: Status: %d", iStatus);
	if (iStatus	!= KErrNone)
		{
		User::Leave(KErrShoutcast_Connecting2Server+iStatus.Int());
		}

	//We use iPtrBuffer2Decode to build the HTTP request
	ResetBufferVars();
	LOG1("SendRequestToServerL: No error, sending HTTP request (path has %d bytes)", iPath.Length());
	//first line of the request
	iPtrBuffer2Decode.Copy(KSCGet);
	iPtrBuffer2Decode.Append(iPath);
	iPtrBuffer2Decode.Append(KSCHttp10);
	//next lines of the request
	iPtrBuffer2Decode.Append(KSCUserAgent);
	iPtrBuffer2Decode.Append(KSCAccept);
	//Indicate to the server we want to receive metadata
	iPtrBuffer2Decode.Append(KSCIcyMetadata);
	iPtrBuffer2Decode.Append(KSCConnectionClose);

	//send it off
	iState = ESendingRequest;
	iSock.Write(iPtrBuffer2Decode, iStatus);
	SetActive();
	TBuf16<1000> zz;
	zz.Copy(iPtrBuffer2Decode);
	LOG1("SendRequestToServerL OK (HTTP request sent:\n%S)->end of HTTP request",&zz);
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ReceiveResponseFromServerL
// Read the response from the server.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ReceiveResponseFromServerL()
	{
	LOG("CShoutcastStream::ReceiveResponseFromServerL: We sent the HTTP request");

	if (iStatus != KErrNone)
		{
		User::Leave(KErrShoutcast_SendingHTTPRequest+iStatus.Int());
		}
	LOG("CShoutcastStream::ReceiveResponseFromServerL: HTTP request sent successfully");

	//receive the HTTP answer
	iState = EReceivingResponse;
	iPtrBuffer2Decode.SetLength(0);
	iSock.Read(iPtrBuffer2Decode,iStatus);
	SetActive();
	LOG("CShoutcastStream::ReceiveResponseFromServerL OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ParseResponseFromServerL
// Parse the response message.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ParseResponseFromServerL()
	{
	TInt pos,len,extraByteNL;
	len = iPtrBuffer2Decode.Size();
	TBool haveError=EFalse;
	LOG2("ParseResponseFromServerL: HTTP answer: we got %d bytes. err = %d",len, iStatus);

	if (iStatus	!= KErrNone)
		{
		if(len==0)
			{
			User::Leave(KErrShoutcast_ReceivingHTTPResponse+iStatus.Int());
			}
		else
			{
			haveError=ETrue;
			}
		}

	//Get and parse the HTTP answer
    if(len < 500 && iStatus != KErrNone)
	 {
        TBuf16<500> zz;
        zz.Copy(iPtrBuffer2Decode);
		LOG1("HTTP answer:\n%S",&zz);
	 };


	//...but first, initialize some data
	ASSERT(iPosBuffer2Decode == 0);
	iTotalBytesReceived+=len;//so, we received some data in the iPtrBuffer2Decode
	iState = EAnswerError;//if we don't get the OK HTTP response, we asume there was an error

	for(;;) // Parse the response header
		{
		iPtrBuffer2Decode.Set(iBuffer2Decode+iPosBuffer2Decode, len-iPosBuffer2Decode, len-iPosBuffer2Decode);
		pos = iPtrBuffer2Decode.Locate(10);
		if(pos==KErrNotFound)
			{
			User::Leave(KErrShoutcast_NoHTTPServer);
			}

		if(pos > 0 && iPtrBuffer2Decode[pos-1] == 13)
			{
			extraByteNL=1;
			}
		else
			{
			extraByteNL=0;
			}

		iPtrBuffer2Decode.Set(iBuffer2Decode+iPosBuffer2Decode,pos-extraByteNL,pos-extraByteNL);
		iPosBuffer2Decode += (pos+1);

		if( pos-extraByteNL == 0 )
			{
			break;//we got to the end of the answer
			}

		//if we are here, we got a line to parse
		//check our line against some usefull header
		if( iPtrBuffer2Decode.Find(KSCOKHeader) != KErrNotFound )
			{
			//we found the OK line
			iState = EAnswerOK;
			LOG("200 OK");
			}
		else if( iPtrBuffer2Decode.Find(KSCContentTypeTag) != KErrNotFound )
			{
			//This tell us which mime type the stream is.
			TInt pos2 = iPtrBuffer2Decode.LocateReverse(':');
			if( pos2 > 0 )
				{
				TPtrC8 contentPtr = iPtrBuffer2Decode.Right(iPtrBuffer2Decode.Length()-pos2-1);
				TBuf<20> contentType;
				contentType.Copy(contentPtr);
				contentType.Trim();
				if ( !contentType.Compare(KSCMimeTypeAudioMpeg) )
					{
					iDataType.Set(KMMFFourCCCodeMP3);
					}
				else if ( !contentType.Compare(KSCMimeTypeAudioAacp) )
					{
					iDataType.Set(KMMFFourCCCodeAAC);
					}
				else if ( !contentType.Compare(KSCMimeTypeAudioAac) )
					{
					iDataType.Set(KMMFFourCCCodeAAC);
					}
				}
			}
		else if( iPtrBuffer2Decode.Find(KSCIcyMetaintTag) != KErrNotFound )
			{
			//we will get MetaData, at some interval. Fill the iMetaint
			TInt pos2 = iPtrBuffer2Decode.LocateReverse(':');
			if( pos2 == KErrNotFound )
				{
				User::Leave(KErrShoutcast_BadIcyMetadataFormat);
				}
			TPtrC8 metaintPtr = iPtrBuffer2Decode.Right(iPtrBuffer2Decode.Length()-pos2-1);
			TBuf8<10> metaint;
			metaint.Copy(metaintPtr);
			metaint.Trim();
			iMetaint = atoi((const char *)metaint.PtrZ());
			iCnt = iMetaint;
			LOG1("We have metadata at %d bytes interval",iMetaint);
			}
		else if( iPtrBuffer2Decode.Find(KSCIcyNameTag) != KErrNotFound )
			{
			LOG("icy-name");
			//this is the server name. Fill the metadataServer
			TInt pos2 = iPtrBuffer2Decode.LocateReverse(':');
			if( pos2 > 0 )
				{
                TInt len = iPtrBuffer2Decode.Length() - pos2 - 1;
                if ( len > iTempMetadata.MaxLength() )
                    {
                    len = iTempMetadata.MaxLength();
                    }
                if ( len > 30 ) //? [Petteri] Old code had a limit of 30, let's observe it...
                    {
                    len = 30;
                    }
                iTempMetadata.Copy(iPtrBuffer2Decode.Mid(pos2+1, len));
                // [Petteri] This is broken if the name is shorter than 30 characters
// 				if ( pos2 + 25 > iTempMetadata.MaxLength() )
// 					{
// 					iTempMetadata.Copy(iPtrBuffer2Decode.Mid(pos2+1, iTempMetadata.MaxLength()-pos2));
// 					}
// 				else
// 					{
// 					iTempMetadata.Copy(iPtrBuffer2Decode.Mid(pos2+1,30));
// 					}
				iMetadata[0]->SetValueL(iTempMetadata);
				} //if pos2==KErrNotFound or simmilar, no big deal, we will not have server name

			}
		else if( iPtrBuffer2Decode.Find(KSCIcyGenreTag) != KErrNotFound )
			{
			LOG("icy-genre");
			//this is the server genre. Fill the metadataGenre
			TInt pos2 = iPtrBuffer2Decode.LocateReverse(':');
			if( pos2 > 0 )
				{
				LOG("icy-genre present");
				//use pos2 as the new length
				pos2 = iPtrBuffer2Decode.Length()-pos2-1;
				if( pos2 > iTempMetadata.MaxLength() )
					{
					pos2 = iTempMetadata.MaxLength();
					}
				iTempMetadata.Copy(iPtrBuffer2Decode.Right(pos2));
				iMetadata[1]->SetValueL(iTempMetadata);
				} //if pos2==KErrNotFound or simmilar, no big deal, we will not have server genre

			}
		else if( iPtrBuffer2Decode.Find(KSCIcyBrTag) != KErrNotFound )
			{
			LOG("icy-br");

			//this is the bitrate
			TInt pos2 = iPtrBuffer2Decode.LocateReverse(':');
			if(pos2 == KErrNotFound)
				{
				User::Leave(KErrShoutcast_BadIcyMetadataFormat);
				}

			TPtrC8 brPtr = iPtrBuffer2Decode.Right(iPtrBuffer2Decode.Length()-pos2-1);
			TBuf8<10> br;
			br.Copy(brPtr);
			br.Trim();
			iBitRate = atoi((const char *)br.PtrZ()) * 1000;
			}
		}

	//check if the answer is EAnswerOK
	if( iState != EAnswerOK || haveError )
		{
		User::Leave(KErrShoutcast_NoShoutcastServer);
		}

	//get the metadata out of the remaining buffer/data
	iLenBuffer2Decode = len-iPosBuffer2Decode;
	GetAndRemoveMetadata(iBuffer2Decode+iPosBuffer2Decode, iLenBuffer2Decode);

	GetAudioSettings();

	//some more metadata
	iTempMetadata.Format(KSCTechDataFormat, iBitRate/1000, iSamplingRate/1000);
	if ( iAudioSettings.iChannels == TMdaAudioDataSettings::EChannelsMono )
		{
		iTempMetadata.Append(KSCMono);
		}
	else
		{
		iTempMetadata.Append(KSCStereo);
		}
	if ( iDataType.FourCC() == KMMFFourCCCodeMP3 )
		{
		iTempMetadata.Append(KSCMp3);
		}
	else if ( iDataType.FourCC() == KMMFFourCCCodeAAC )
		{
		iTempMetadata.Append(KSCAacp);
		}
	iMetadata[2]->SetValueL(iTempMetadata);

	iState = EInitStreamOutput;
	// Notify the client that metadata is available
	iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),0x01FF);

	// Initialize the output stream
	InitStreamOutputL();
	//done
	LOG("ParseResponseFromServerL OK");
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::GetAudioSettings
// -----------------------------------------------------------------------------
//
void CShoutcastStream::GetAudioSettings()
	{

	if ( ScanHeader(iBuffer2Decode, iLenBuffer2Decode) )
		{
		switch ( iSamplingRate )
			{
			case 96000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate96000Hz;
				break;
			case 64000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate64000Hz;
				break;
			case 48000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate48000Hz;
				break;
			case 44100:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate44100Hz;
				break;
			case 32000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate32000Hz;
				break;
			case 24000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate24000Hz;
				break;
			case 22050:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate22050Hz;
				break;
			case 16000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate16000Hz;
				break;
			case 12000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate12000Hz;
				break;
			case 11025:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate11025Hz;
				break;
			case 8000:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate8000Hz;
				break;
			default:
				iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate8000Hz;
			}

		if ( iChannels >= 2 )
			{
			iAudioSettings.iChannels = TMdaAudioDataSettings::EChannelsStereo;
			}
		else
			{
			iAudioSettings.iChannels = TMdaAudioDataSettings::EChannelsMono;
			}
		}
	else
		{
		// In the event that scan header didn't find any valid encoding data we have to derive them
		// from the bitrate received from the server.
		DeriveAudioSettings();
		}

	}

// -----------------------------------------------------------------------------
// CShoutcastStream::DeriveAudioSettings
// This is only needed if we are unable to extract audio settings from the stream.
// These assumptions are not always accurate.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::DeriveAudioSettings()
	{

	if ( iBitRate >= 64000)
		{
		// Typical sampling rate for these high bitrates
		iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate44100Hz;
		iAudioSettings.iChannels = TMdaAudioDataSettings::EChannelsStereo;
		iSamplingRate = 44100;
		}
	else // lower bitrates
		{
		if ( iDataType.FourCC() == KMMFFourCCCodeAAC )
			{
			// Typical sampling rate for AAC+ stream
			iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate24000Hz;
			iAudioSettings.iChannels = TMdaAudioDataSettings::EChannelsStereo;
			iSamplingRate = 24000;
			}
		else
			{
			// Typical sampling rate for low bitrates mp3 stream
			iAudioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate16000Hz;
			iAudioSettings.iChannels = TMdaAudioDataSettings::EChannelsMono;
			iSamplingRate = 16000;
			}
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::PrimeL
//
//   -allocates the iBuffer2Decode buffer
//   -allocate metadata entries
//   -opens some stuff and gets some data in iBuffer2Decode buffer
//   -instantiates the mp3 codec,
//   -gets the audio parameter
//   -allocates the iBuffer2Play buffer
//   -negotiates with the sink
//   -initializes some of the metadata entries
// -----------------------------------------------------------------------------
//
void CShoutcastStream::PrimeL()
	{
	LOG("CShoutcastStream::PrimeL");
	//state should be ENotConnected
	if ( iState != ENotConnected )
		{
		User::Leave(KErrNotSupported);
		}

	//-allocate the iBuffer2Decode buffer
	iBuffer2Decode = new (ELeave) TUint8[BUFFER2DECODE_SIZE];
	// allocate the play buffer
	iBuffer2Play = new (ELeave) TUint8[BUFFER2PLAY_SIZE];
	iMetadataBuffer = new (ELeave) TUint8[METADATA_BUFFER_SIZE];

	ResetBufferVars();

	//-add metadata fields. Intended: server name, server genre, artist, song, bitrate (constant) + kHz + stereo, bytes by now + cost (total: 6)
	CMetaDataEntry *mmfMetadata;
	iTempMetadata.Copy(_L(""));

	//add metadata - server name
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaServer, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - server genre
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaGenre, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - technical details
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaTechDetails,iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - artist + song
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaArtistTitle, iTempMetadata);
	iMetadata.Append(mmfMetadata);

    //add metadata - bytes
	iTempMetadata.Format(_L("%dMB"),0);
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaBytes, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - price
	iTempMetadata.Format(_L("price: %d"),0);
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaPrice, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - buffer fill
	iTempMetadata.Format(_L("%d"),0);
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaBuffer, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	//add metadata - bitrate
	iTempMetadata.Format(_L("%d/%d"),0);
	mmfMetadata = CMetaDataEntry::NewL(KSCMetaBitrate, iTempMetadata);
	iMetadata.Append(mmfMetadata);

	// try to establish connection with the server
	ConnectToServerL();
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::PlayL
// Start playing the stream by filling the buffer to be played and write it to
// the stream.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::PlayL()
	{
	LOG("PlayL");
	iPausedForBuffering = EFalse;
	//update metadata on the client
	FillBufferL();
	//reset the bitrate stuff
	iLastReading = 0;
	iBitsRead.Reset();
	iTimeIntervals.Reset();
	LOG("PlayL OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::Pause
// Pauses the playback but the engine continues to receive data from the
// network. The stream output is stopped.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::Pause()
	{
	iStreamOutput->Stop();
	LOG("PauseL, OK");
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::Stop
// Stops the playback. The engine stops receiving data from the
// network. The stream output is stopped.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::Stop()
	{
	LOG("StopL");
	if ( iState == EData )
		{
		iStreamOutput->Stop();
		}
	//we have to undo all that the PrimeL function did
	Cancel();//make sure we do not have any request pending
	CloseAndClean();
	//we expect that Prime & Play to be called in order to play again
	LOG("StopL OK");
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::DoCancel
// Cancels any outstanding requests
// -----------------------------------------------------------------------------
//
void CShoutcastStream::DoCancel()
	{
	LOG("DoCancel was called!!!");
	iSock.CancelAll();
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::RunL
// Invoked when one of these requests completes.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::RunL()
	{
	switch ( iState )
		{
		case EConnecting:
			SendRequestToServerL();
			break;
		case ESendingRequest:
			ReceiveResponseFromServerL();
			break;
		case EReceivingResponse:
			ParseResponseFromServerL();
			break;
		case EData:
			{
			if ( iStatus != KErrNone )
				{
				CloseAndClean();
				Disconnect(iStatus.Int());
				return;
				}
			if ( ReadRequestHelperDone() != KErrNone )
				{
				CloseAndClean();
				Disconnect(iStatus.Int());
				return;
				}
			};
			break;
		default:
			ASSERT(0);
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::RunError
// Called when the RunL leaves.
// If we are here, it means that we failed to connect somehow!
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::RunError(
	TInt aError )
	{
	//send the error as an event to the client
	iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), aError);
	return KErrNone;
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::GetNumberOfMetaDataEntries
// Get number of metadata entries
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::GetNumberOfMetaDataEntries()
	{
	return NR_METADATA_FIELDS;
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::GetNumberOfMetaDataEntries
// Get a particular metadata entry.
// -----------------------------------------------------------------------------
//
CMetaDataEntry* CShoutcastStream::GetMetaDataEntryL(
	TInt aIndex )
	{
	//check if this is "special" metadata

	LOG("GetMetaDataEntryL");

	if( aIndex == KShoutcastMetaData_URL )
		{
		CMetaDataEntry *mmfMetadata;
		iTempMetadata.Copy(iURL);
		mmfMetadata = CMetaDataEntry::NewL(KSCMetaServer, iTempMetadata);
		return mmfMetadata;
		};

	//this is normal metadata
	if ( aIndex >= iMetadata.Count() )
		{
		return NULL;
		}
	else
		{
		CMetaDataEntry* md = iMetadata[aIndex];
		md = CMetaDataEntry::NewL(*md);
		return md;
		}
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::ReadRequest
// The read request is done when all the requested data has been read.
// The read request is made intelligently:
// 		-first, we read until the metadata count byte.
//		-if we have metadata, then we read metadata (in a separate buffer),
//		-then we read what is left
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ReadRequest()
{
	TInt data2read = DATA2READ;
	TUint8 *buffer2read = NULL;

	if ( iReadData == -1)
		{
		//case 1: the buffer is partially filled (data is in the middle)
		//we have to read from iBuffer2Decode+iPosBuffer2Decode+iLenBuffer2Decode
		//... 8k (DATA2READ)
		//... or until the end of the buffer
		if ( data2read > BUFFER2DECODE_SIZE-iPosBuffer2Decode-iLenBuffer2Decode )
			{
			data2read = BUFFER2DECODE_SIZE-iPosBuffer2Decode-iLenBuffer2Decode;
			}
		//check for transition to case 2
		if ( data2read < READ_EPSILON && iPosBuffer2Decode>iMaxFrameSize )
			{
			//go to case 2
			iReadData = 0;
			LOG2("RR transition case 1->case 2 (pos=%d, len=%d)",iPosBuffer2Decode,iLenBuffer2Decode);
			ReadRequest();
			return;
			}
		else
			{
			//stick to case 1
			buffer2read = iBuffer2Decode+iPosBuffer2Decode+iLenBuffer2Decode;
			LOG1("RR case 1: have to read %d bytes",data2read);
			}
		}
	else
		{
		//case 2: we fill the buffer from the beginning (or from the middle)
		//we have to read from iBuffer2Decode+iMaxFrameSize+iReadData
		//... 8k (DATA2READ)
		//... or until the data (iPosBuffer2Decode)
		ASSERT(iReadData >= 0);
		ASSERT(iMaxFrameSize >= 0);
		ASSERT(iMaxFrameSize < 3000);
		ASSERT(iPosBuffer2Decode >= iMaxFrameSize);
		ASSERT(iMaxFrameSize + iReadData <= iPosBuffer2Decode);
		if ( iMaxFrameSize+iReadData+data2read > iPosBuffer2Decode )
			{
			data2read = iPosBuffer2Decode-iMaxFrameSize-iReadData;
			}
		buffer2read = iBuffer2Decode+iMaxFrameSize+iReadData;
		LOG1("RR case 2: have to read %d bytes",data2read);
		}

	if ( data2read < READ_EPSILON )
		{
		//the buffer looks completely full!
		// suspend the read
		iReadingActive = EFalse;
		LOG("RR OK (buffer full, no reading)");

		if ( iPausedForBuffering ) // in case we have paused playback for buffering...
			{
			TRAPD(err, PlayL());
			if ( err )
				{
				iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), err);
				}
			else
				{
				// send buffering complete event!!
				iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),KShoutcastEvent_BufferingComplete);
				}
			}
		}
	else
		{
		//we need to read data!
		ASSERT(data2read > 0);
		iPtrBuffer2Decode.Set(buffer2read,0,data2read);
		iReadingActive = ETrue;
		iRRreq = data2read;
		iRRread = 0;
		LOG1("RR OK (will read %d bytes)",data2read);
		ReadRequestHelper();
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ReadRequestHelper
// Read the requested amount of data from the socket. When the asynchronous read
// completes, our RunL will be called.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::ReadRequestHelper()
	{
	//read from the socket
	iSock.Read(iPtrBuffer2Decode,iStatus);
	SetActive();
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ReadRequestHelperDone
// The read request is completed. Extract metadata from the buffer if there's
// any and request additional reads if the buffer wasn't full.
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::ReadRequestHelperDone()
	{
	TInt readSize = iPtrBuffer2Decode.Size();
	TInt maxSize = iPtrBuffer2Decode.MaxSize();
	LOG2("RRHD read %d (status: %d)",readSize,iStatus);

	GetAndRemoveMetadata(const_cast<TUint8*>(iPtrBuffer2Decode.Ptr()),readSize);

	iRRread+=readSize;

	//check if we have read the full length
	if ( maxSize-readSize<5 || (readSize==0 && iStatus.Int()==0) )
		{
		//this reading is done!
		return ReadRequestDone(iRRreq,iRRread);
		}

	LOG2("RRHD: Partial read of %d, requesting again %d bytes",readSize,maxSize-readSize);
	iPtrBuffer2Decode.Set(const_cast<TUint8*>(iPtrBuffer2Decode.Ptr())+readSize,0,maxSize-readSize);
	ReadRequestHelper();

	return KErrNone;
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::ReadRequestDone
// The entire read request buffer is full. Now process the contents and re-new
// the read request.
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::ReadRequestDone(
	TInt aRequested,
	TInt aRead )
	{
	TInt bufferFill;
	TInt instantBitrate = 0;
	TInt avgBitrate = 0;
	TInt metadataInd = 0;

	//if we are here, it means that we read normal data
	LOG2("RRD: We have read normal data (req=%d, read=%d)",aRequested,aRead);
	iTotalBytesReceived += aRead;

	// 100kb: 1024*100 = 102400 (decimal)
	// 1MB: 1048576 binary bytes
	// In order to get exactly 0.1 MB, and for easy calculation, the constant 104857 is used here.

	//every 100kB we signal metadata modifications
	if( (iNrHKB+1) * K100kBFactor < iTotalBytesReceived )
		{
		iNrHKB++;
		//some more metadata
		iTempMetadata.Format(KSCDataRcvdFormat, iTotalBytesReceived/K1MBFactor);
		// It's ok if we can't set the value.
		TRAPD(err,
			{
			iMetadata[4]->SetValueL(iTempMetadata);
			iTempMetadata.Format(KSCMetaPriceFormat, iTotalBytesReceived/K1MBFactor * PRICE_PER_MB);
		    iMetadata[5]->SetValueL(iTempMetadata);
			});

		if ( !err )
			{
			// Indicate which metadata was updated
			metadataInd = 0x0130;
			}
		}

	//do something with the data we just read!
	if( iReadData == -1)
		{
		//case 1: we should increase the data length
		iLenBuffer2Decode+=aRead;
		bufferFill=(TInt)(100.0*iLenBuffer2Decode/(BUFFER2DECODE_SIZE-iMaxFrameSize));
		}
	else
		{
		//case 2: we need to increase the iReadData
		iReadData += aRead;
		bufferFill = (TInt)(100.0*(iLenBuffer2Decode+iReadData)/(BUFFER2DECODE_SIZE-iMaxFrameSize));
		}

	//compute bitrate
	if( iLastReading.Int64() == 0)
		{
		iLastReading.UniversalTime();
		}
	else
		{
		TInt i,tbits;
		TTime now;
		TInt64 avg;
		now.UniversalTime();
		TInt64 diff=now.Int64()-iLastReading.Int64();
		TInt bits=(TInt)(aRead*7812.5);
        if ( diff > 0 )
            {
            instantBitrate=(TInt64(bits)/diff).GetTInt();//7812.5=8*1000*1000/1024 (8=bits, 1024=kbits, 100000 from microseconds)
            iPrevInstantBitrate = instantBitrate;
            }
        else
            {
            instantBitrate = iPrevInstantBitrate;
            }
		//compute now the averaged bitrate
		iBitsRead.Append(bits);
		iTimeIntervals.Append(diff);
		ASSERT(iBitsRead.Count()==iTimeIntervals.Count());
		tbits=iBitsRead[0];
		avg=iTimeIntervals[0];
		for( i = 1; i < iBitsRead.Count(); i++)
			{
			tbits += iBitsRead[i];
			avg += iTimeIntervals[i];
			};
        if ( avg > 0 )
            {
            avgBitrate=(TInt64(tbits)/avg).GetTInt();
            iPrevAvgBitrate = avgBitrate;
            }
        else
            {
            avbBitrate = iPrevAvgBitrate;
            }
		if( iBitsRead.Count() > BITRATE_SAVE_COUNT )
			{
			iBitsRead.Remove(0);
			iTimeIntervals.Remove(0);
			}
		iLastReading=now;
		}

	//fill buffering metadata
	if ( bufferFill > 100)
		{
		bufferFill=100;//this can happen when a part of iMaxFrameSize is also filled
		}

	TRAPD(err,
		{
		iTempMetadata.Format(KSCBufFormat, bufferFill);
		iMetadata[6]->SetValueL(iTempMetadata);
        if ( (avgBitrate > 0) && (instantBitrate > 0) )
            {
            iTempMetadata.Format(KSCMetaBitrateFormat, avgBitrate, instantBitrate);
            }
		iMetadata[7]->SetValueL(iTempMetadata);
		});

	if ( !err )
		{
		// Indicate which metadata was updated
		metadataInd = metadataInd | 0x01C0;
		}

	// Send metadata update event to client
	iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid), metadataInd);

	//if we are here, it means that we have finished reading our data!!
	LOG1("RRD: Done reading %d bytes",aRead);

	//Read again!
	ReadRequest();

	return KErrNone;
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::RemoveMetadata
// Remove metadata from the buffer
// -----------------------------------------------------------------------------
//
void CShoutcastStream::RemoveMetadata(
	TDes8& ptrMetadata )
	{
	ASSERT(iMetaint > 0);//we have metadata

	TInt pos1,pos2;
	pos1 = ptrMetadata.Find(KSCStreamTitle);
	pos2 = ptrMetadata.Find(_L8("';"));

	if ( pos1 != KErrNotFound && pos2 != KErrNotFound )
		{
		//use pos2 as length
		pos2 = pos2-pos1-13;//13 is the number of characters in "StreamTitle='"
		if ( pos2 > iTempMetadata.MaxLength() )
			{
			pos2=iTempMetadata.MaxLength();
			}
		iTempMetadata.Copy(ptrMetadata.Mid(pos1+13,pos2));
		TRAPD(err, iMetadata[3]->SetValueL(iTempMetadata));
		LOG2("MD: Metadata(%d):(%S)", pos2, &iTempMetadata);
		// Send metadata event to the client
		if ( !err )
			{
			iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),0x0108);
			}
		}
	else
		{
		LOG("MD: Metadata found but Title NOT FOUND !!! ");
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::GetAndRemoveMetadata
// Get the metadata and remove it from the data stream
// -----------------------------------------------------------------------------
//
void CShoutcastStream::GetAndRemoveMetadata(
	TUint8 *aBuf,
	TInt &aLength )
	{
	if ( iMetaint == 0 )
		{
		return; //we do not have metadata!
		}

	TInt dataLength = aLength;
	TInt pos = 0;

	while ( iCnt < dataLength )
		{
		dataLength -= iCnt;
		pos += iCnt;
		//we have metadata here!
		iDataSeen = 16 * aBuf[pos];
		LOG1("MD: Found Metadata %d bytes",iDataSeen);
		if ( iDataSeen )
			if ( iDataSeen < METADATA_BUFFER_SIZE )
				{
				//we really have metadata!
				TPtr8 ptrMetadata(aBuf+pos+1,iDataSeen,iDataSeen);
				RemoveMetadata(ptrMetadata);
				}
			else
				{
				//we lost the sync
				ASSERT(0);
				}
		//remove the metadata
		iDataSeen++;
		LOG2("Removing metadata: pos=%d, removed=%d",pos,iDataSeen);
		memmove(aBuf+pos,aBuf+pos+iDataSeen,aLength-pos-iDataSeen);
		iCnt = iMetaint;
		aLength -= iDataSeen;
		dataLength -= iDataSeen;
		LOG1("After memmove: aLength=%d",aLength);
		}
	iCnt -= dataLength;
	LOG2("MD: we exit with %d bytes (cnt is %d)",aLength,iCnt);
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::ScanHeader
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::ScanHeader(
	TUint8 *aBuf,
	TInt aLength )
	{

	TInt length = 0;
    TInt readLen = aLength;
    TInt seekLen;
    const TUint8* seekPtr;
    TInt seekOffset;

    TAudioFrameInfo frameInfo;

    if ( iDataType.FourCC() == KMMFFourCCCodeMP3 )
    	{
		seekOffset = Mp3SeekSync(aBuf, aLength);
		seekLen = readLen - seekOffset;
		seekPtr = aBuf + seekOffset;
		length = GetMp3FrameInfo(seekPtr, seekLen, frameInfo);
		}
	else if ( iDataType.FourCC() == KMMFFourCCCodeAAC )
		{
		seekOffset = AacSeekSync(aBuf, aLength);
		seekLen = readLen - seekOffset;
		seekPtr = aBuf + seekOffset;
		length = GetAacFrameInfo(seekPtr, seekLen, frameInfo);
		}

	if ( length > 0  &&  frameInfo.iBitRate > 0 )
		{
		iSamplingRate = frameInfo.iSamplingRate;
		iBitRate = frameInfo.iBitRate;
		iChannels = frameInfo.iChannels;
		return ETrue;
		}
	else
		{
		return EFalse;
		}
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::Mp3SeekSync
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::Mp3SeekSync(
	TUint8 *aBuf,
	TInt aLength )
    {
    const TUint8* buf = aBuf;
    const TInt KMaxFrames = 3;          // number of frames to check
    const TInt KNotFound = aLength;     // sync not found position
    TAudioFrameInfo frameInfo;           // frame parameters
    TInt i = 0;
    TInt syncPos = KNotFound;
    TInt maxSeek = KMaxFrames;
    const TUint8* endPtr = buf + aLength;

    // Seek a valid frame candidate byte by byte until a valid frame
    // is found or all bytes have been checked.
    while ( buf < endPtr  &&  syncPos == KNotFound )
	    {
        TInt seekCount = 0;
        const TUint8* framePtr = buf;
        TInt frameBufLen = aLength;
        syncPos = i;

		// Check the validity of this frame candidate and the nearest next
        // frames. If they are not OK, syncPos will be set to KNotFound.
        while ( framePtr < endPtr  &&  syncPos != KNotFound  && seekCount < maxSeek )
	        {
            TInt length = GetMp3FrameInfo(framePtr, frameBufLen, frameInfo);

			if ( length == 0 )
            	{
                syncPos = KNotFound;
				}
            if ( (length > 0) && (frameInfo.iBitRate < 0 ) )
            	{
                maxSeek = 1; // free formatcase
				}
            framePtr += length;
            frameBufLen -= length;
            seekCount++;

			// consider SYNC not found if we reach end of buffer before finding 3 SYNC frames
			if ((framePtr >= endPtr) && (seekCount < maxSeek))
				{
				syncPos = KNotFound;
				aBuf += (aLength - 1);      // force an exit from while loop
				}

        	}
        buf++; aLength--; i++;
    	}
    return syncPos;
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::GetMp3FrameInfo
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::GetMp3FrameInfo(
	const TUint8* aBuf,
	TInt aBufLen,
	TAudioFrameInfo& aInfo )
    {
    TInt length = 0;
    TUint temp;
    TUint lTempVal;

	if ( aBufLen >= KMp3FrameHeaderSize )
	    {
		// Extract header fields to aInfo and check their bit syntax
		// (including the sync word!). If the syntax is not OK the length
		// is set to zero.

		temp = 0;
		temp = aBuf[0] << 24;
		temp |= (aBuf[1] << 16);
		temp |= (aBuf[2] << 8);
		temp |= aBuf[3];
		if ( ((temp >> 21) & 0x7FF) != 0x7FF )
			{
			return length;
			}

		lTempVal = (temp >> 19) & 0x03;
		switch ( lTempVal )
			{
			case 0:
				aInfo.iId = 2;  // MPEG2.5
				aInfo.iFrameSamples = 576;
				break;
			case 1:
				return length;
			case 2:
				aInfo.iId = 0;  // MPEG 2
				aInfo.iFrameSamples = 576;
				break;
			case 3:
				aInfo.iId = 1;  // MPEG 1
				aInfo.iFrameSamples = 1152;
				break;
			}

		lTempVal = (temp >> 17) & 0x03;
		if ( lTempVal != 1 )
			{
			return length;
			}

		lTempVal = (temp >> 12) & 0x0F;
		aInfo.iBitRate = KMp3BitRateTable[aInfo.iId][lTempVal] * 1000;

		if ( aInfo.iBitRate == 0 )
			{
			return length;
			}

		lTempVal = (temp >> 10) & 0x03;
		if ( lTempVal == 3 )
			{
			return length;
			}
		else
			{
			aInfo.iSamplingRate = KMp3SamplingFrequencyTable[aInfo.iId][lTempVal];
			}

		aInfo.iPadding = (temp >> 9) & 0x01;

		lTempVal = (temp >> 6) & 0x03;
		aInfo.iMode = lTempVal;

		if ( lTempVal == 3 )
			{
			aInfo.iChannels = 1;
			}
		else
            {
            aInfo.iChannels = 2;
            }

        aInfo.iSamplingRateOut = aInfo.iSamplingRate;
        aInfo.iChannelsOut = 2; /* always set to stereo output */

        aInfo.iFrameSamplesOut = aInfo.iFrameSamples;

		if ( aInfo.iBitRate == -1 )
			{
			// For free mode operation
			length = KMp3MaxFrameSize;
			}

		if ( aInfo.iSamplingRate > 0  &&  aInfo.iBitRate > 0 )
			{
			length = (144 * aInfo.iBitRate) / aInfo.iSamplingRate;

			if ( aInfo.iId != 1 )
				{
				length >>= 1; /*for MPEG2 and MPEG2.5 */
				}

			if ( aInfo.iPadding )
				{
				length++;
				}
			}

		aInfo.iFrameSize = length;

    	}
    return length;
    }


// -----------------------------------------------------------------------------
// CShoutcastStream::AacSeekSync
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::AacSeekSync(
	TUint8 *aBuf,
	TInt aLength )
    {
    TInt seekLen = 0;
    while ( seekLen < (aLength - 4) )
        {
        if ( IsValidAacFrame(aBuf + seekLen, aLength - seekLen) )
        	{
            return seekLen;
			}
        seekLen++;
        }
    return aLength;
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::IsValidAacFrame
// -----------------------------------------------------------------------------
//
TBool CShoutcastStream::IsValidAacFrame(
	const TUint8* aPtr,
	TInt aBufLen )
    {
	if ( (aPtr[0] == 'A') && (aPtr[1] == 'D') &&
		 (aPtr[2] == 'I') && (aPtr[3] == 'F') )
		{
		return ETrue;
		}

	if ( (aPtr[0] == 0xFF) && ((aPtr[1] >> 4) == 0xF) && // Check syncword FFF
         ((aPtr[1] >> 1) & 0x3) == 0 // Check layer == 0
         )
        {
        TInt frameLen = 0;
        frameLen = (aPtr[3] & 0x3) << 11;
        frameLen |= aPtr[4] << 3;
        frameLen |= ((aPtr[5] >> 5) & 0x7);

        if ( !frameLen || frameLen > (aBufLen-4) )
           {
           return EFalse;
           }

        // Check next frame
        const TUint8* nextFramePtr = aPtr + frameLen;
	    if ( ((nextFramePtr[0] == 0xFF) && ((nextFramePtr[1] >> 4) == 0xF)) && // Check syncword FFF
             ((nextFramePtr[1] >> 1) & 0x3) == 0 // Check layer == 0
             )
            {
            return ETrue;
            }
        }
    return EFalse;
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::GetAacFrameInfo
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::GetAacFrameInfo(
	const TUint8* aBuf,
	TInt aBufLen,
	TAudioFrameInfo& aInfo )
    {
    TInt length = 0;
    TUint temp;
    TUint lTempVal;
    TUint lTemp1;
	TUint lUnSupported = 0;

	if ( aBufLen >= KAacFrameHeaderSize )
	    {
        // Extract header fields to aInfo and check their bit syntax
        // (including the sync word!). If the syntax is not OK the length
        // is set to zero.

        lTemp1 = aBuf[0] << 24;
        lTemp1 |= (aBuf[1] << 16);
        lTemp1 |= (aBuf[2] << 8);
        lTemp1 |= aBuf[3];
		temp = (lTemp1 >> 20) & 0xFFF;

		if((temp & 0xFFF) == 0xFFF)
        	{
			// ADTS Frame
			temp = (lTemp1 >> 19) & 0x01;
			aInfo.iId = temp;  // 0 for MPEG 4 and 1 for MPEG 2

            // Sampling Frequency Index
			temp = (lTemp1 >> 10) & 0x0F;

			if(temp > 11)
				{
				lUnSupported = 1;
				}
			else
				{
				aInfo.iSamplingRate = KAacSamplingFrequencyTable[temp];
				}

			// The header doesn't tell us about bitrate so use the bitrate from the server
			aInfo.iBitRate = iBitRate;

            // Channel Configuration
			lTempVal = (lTemp1 >> 6) & 0x07;

			if(lTempVal > 2)
				{
				lUnSupported = 1;
				aInfo.iChannels = lTempVal;
				}
			else if (lTempVal == 0 )
				{
				aInfo.iChannels = 2;
				}
			else
				{
				aInfo.iChannels = lTempVal;
				}

			// frame_length
			length = (lTemp1 & 0x03) << 11;

            lTemp1 = 0;
            lTemp1 = aBuf[4] << 24;
            lTemp1 |= (aBuf[5] << 16);
            lTemp1 |= (aBuf[6] << 8);

			/* adts_buffer_fullness */
			temp = temp | ((lTemp1 >> 21) & 0x7FF);

			/*Number of raw data blocks*/
			TInt numberOfRawDataBlks = ((lTemp1 >> 8) &0x03) + 1;

            if(temp > (1536 * numberOfRawDataBlks))
            	{
				/* 1536 bytes per raw datablock in the ADTS frame */
				lUnSupported = 1;
            	}

			/* Update Framesize only if the bit stream is supported */
			if(!lUnSupported)
				{
				aInfo.iFrameSize = temp;
				length = temp;
				}
        	}
    	}
    return length;
    }

// -----------------------------------------------------------------------------
// CShoutcastStream::GetIPFromURL
// Get the IP address from the URL
// -----------------------------------------------------------------------------
//
void CShoutcastStream::GetIPFromURL()
 	{
	//parse the iURL and get the IP:port
	//the URL is iURL. The target IP:port is *addr
	//we must keep the iURL constant
	TUint32 ip=0;
	TUint port=80;//the default port

	TInt length,start,end,i;
	TBuf8<URL_SIZE> url;


	//assume the http:// in front of the string
	ASSERT(iURL.Find(_L8("http://"))==0);
	start=7;
	//get the length of the url
	length=iURL.Length();

	//get the url without the http://
	url=iURL.Right(length-start);

	length = url.Find(_L8(":"));
	// get the ip addr from the url
	if ( length == KErrNotFound )
		{
		iIp = url;
		}
	else
		{
		iIp = url.Left(length);
		}

	if ( url[0] >= 48 && url[0] <=57)
		{
		//the IP starts with a number, so we have a numeric IP
		for( i = 3; i>=0; i-- )
			{
			ip = ip*256+(unsigned)atoi((const char *)url.PtrZ());
			LOG1("GetIPFromURL: ip: %u",(unsigned)atoi((const char *)url.PtrZ()));
			//get the next dot/:
			if ( i )
				{
				end=url.Locate('.');
				url.Delete(0,end+1);
				}
			}
		}
	else
		{
		//the IP does not start with a number!
		ASSERT(0);//### not handled yet
		};
	//get the port, if any
	end=url.Locate(':');
	if ( end != KErrNotFound )
		{
		//we found a port!
		url.Delete(0,end+1);
		port = (unsigned)atoi((const char *)url.PtrZ());
		}

	//get the iPath, this is what remains
	end = url.Locate('/');
	if ( end != KErrNotFound )
		{
		//we have a path
		url.Delete(0,end);//we do not delete the /
		iPath.Copy(url);
		}
	else
		{
		//there is no path! we make it /
		iPath.Copy(_L8("/"));
		}

	iAddr = new TInetAddr(ip,port);

	}

// -----------------------------------------------------------------------------
// CShoutcastStream::Disconnect
// Send disconnect event to client.
// -----------------------------------------------------------------------------
//
void CShoutcastStream::Disconnect(
	TInt aError)
	{
	LOG1("In Disconnect (status: %d)", aError);
	iDispatcher->SendEvent(TUid::Uid(KShoutcastStreamUid),KShoutcastEvent_Disconnected);
	}


// -----------------------------------------------------------------------------
// CShoutcastStream::SetVolume
// Sets the volume level
// -----------------------------------------------------------------------------
//
void CShoutcastStream::SetVolume(
	TInt aVolume )
	{
	iVolume = aVolume;
	if ( iStreamOutput )
		{
		iStreamOutput->SetVolume(aVolume);
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::Volume
// Get the volume level
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::Volume()
	{
	if ( iStreamOutput )
		{
		return iStreamOutput->Volume();
		}
	else
		{
		return iVolume;
		}
	}

// -----------------------------------------------------------------------------
// CShoutcastStream::MaxVolume
// Get the maximum volume level
// -----------------------------------------------------------------------------
//
TInt CShoutcastStream::MaxVolume()
	{
	if ( iStreamOutput )
		{
		return iStreamOutput->MaxVolume();
		}
	else
		{
		return 10;
		}
	}

// End of File
